/*
    This file is part of Magnum.

    Copyright © 2010, 2011, 2012, 2013, 2014, 2015
              Vladimír Vondruš <mosra@centrum.cz>

    Permission is hereby granted, free of charge, to any person obtaining a
    copy of this software and associated documentation files (the "Software"),
    to deal in the Software without restriction, including without limitation
    the rights to use, copy, modify, merge, publish, distribute, sublicense,
    and/or sell copies of the Software, and to permit persons to whom the
    Software is furnished to do so, subject to the following conditions:

    The above copyright notice and this permission notice shall be included
    in all copies or substantial portions of the Software.

    THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
    IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
    FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
    THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
    LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
    FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER
    DEALINGS IN THE SOFTWARE.
*/

#include "AbstractQuery.h"

#include <Corrade/Utility/Assert.h>

#include "Magnum/Context.h"
#include "Magnum/Implementation/DebugState.h"
#include "Magnum/Implementation/QueryState.h"
#include "Magnum/Implementation/State.h"

namespace Magnum {

AbstractQuery::AbstractQuery(GLenum target): _target{target} {
    (this->*Context::current()->state().query->createImplementation)();
}

#ifdef MAGNUM_BUILD_DEPRECATED
AbstractQuery::AbstractQuery(): _target{} {
    createImplementationDefault();
}
#endif

AbstractQuery::~AbstractQuery() {
    /* Moved out, nothing to do */
    if(!_id) return;

    #ifndef MAGNUM_TARGET_GLES2
    glDeleteQueries(1, &_id);
    #elif !defined(CORRADE_TARGET_EMSCRIPTEN)
    glDeleteQueriesEXT(1, &_id);
    #else
    CORRADE_ASSERT_UNREACHABLE();
    #endif
}

void AbstractQuery::createImplementationDefault() {
    #ifndef MAGNUM_TARGET_GLES2
    glGenQueries(1, &_id);
    #elif !defined(CORRADE_TARGET_EMSCRIPTEN)
    glGenQueriesEXT(1, &_id);
    #else
    CORRADE_ASSERT_UNREACHABLE();
    #endif
}

#ifndef MAGNUM_TARGET_GLES
void AbstractQuery::createImplementationDSA() {
    glCreateQueries(_target, 1, &_id);
}
#endif

std::string AbstractQuery::label() const {
    #ifndef MAGNUM_TARGET_GLES
    return Context::current()->state().debug->getLabelImplementation(GL_QUERY, _id);
    #else
    return Context::current()->state().debug->getLabelImplementation(GL_QUERY_KHR, _id);
    #endif
}

AbstractQuery& AbstractQuery::setLabelInternal(const Containers::ArrayReference<const char> label) {
    #ifndef MAGNUM_TARGET_GLES
    Context::current()->state().debug->labelImplementation(GL_QUERY, _id, label);
    #else
    Context::current()->state().debug->labelImplementation(GL_QUERY_KHR, _id, label);
    #endif
    return *this;
}

bool AbstractQuery::resultAvailable() {
    GLuint result;
    #ifndef MAGNUM_TARGET_GLES2
    glGetQueryObjectuiv(_id, GL_QUERY_RESULT_AVAILABLE, &result);
    #elif !defined(CORRADE_TARGET_EMSCRIPTEN)
    glGetQueryObjectuivEXT(_id, GL_QUERY_RESULT_AVAILABLE_EXT, &result);
    #else
    CORRADE_ASSERT_UNREACHABLE();
    #endif
    return result == GL_TRUE;
}

#ifndef DOXYGEN_GENERATING_OUTPUT
template<> UnsignedInt AbstractQuery::result<UnsignedInt>() {
    UnsignedInt result;
    #ifndef MAGNUM_TARGET_GLES2
    glGetQueryObjectuiv(_id, GL_QUERY_RESULT, &result);
    #elif !defined(CORRADE_TARGET_EMSCRIPTEN)
    glGetQueryObjectuivEXT(_id, GL_QUERY_RESULT_EXT, &result);
    #else
    CORRADE_ASSERT_UNREACHABLE();
    #endif
    return result;
}

template<> bool AbstractQuery::result<bool>() { return result<UnsignedInt>() != 0; }

template<> Int AbstractQuery::result<Int>() {
    Int result;
    #ifndef MAGNUM_TARGET_GLES
    glGetQueryObjectiv(_id, GL_QUERY_RESULT, &result);
    #elif !defined(CORRADE_TARGET_EMSCRIPTEN) && !defined(CORRADE_TARGET_NACL)
    glGetQueryObjectivEXT(_id, GL_QUERY_RESULT_EXT, &result);
    #else
    CORRADE_ASSERT_UNREACHABLE();
    #endif
    return result;
}

#ifndef MAGNUM_TARGET_WEBGL
template<> UnsignedLong AbstractQuery::result<UnsignedLong>() {
    UnsignedLong result;
    #ifndef MAGNUM_TARGET_GLES
    glGetQueryObjectui64v(_id, GL_QUERY_RESULT, &result);
    #elif !defined(CORRADE_TARGET_EMSCRIPTEN) && !defined(CORRADE_TARGET_NACL)
    glGetQueryObjectui64vEXT(_id, GL_QUERY_RESULT_EXT, &result);
    #else
    CORRADE_ASSERT_UNREACHABLE();
    #endif
    return result;
}

template<> Long AbstractQuery::result<Long>() {
    Long result;
    #ifndef MAGNUM_TARGET_GLES
    glGetQueryObjecti64v(_id, GL_QUERY_RESULT, &result);
    #elif !defined(CORRADE_TARGET_EMSCRIPTEN) && !defined(CORRADE_TARGET_NACL)
    glGetQueryObjecti64vEXT(_id, GL_QUERY_RESULT_EXT, &result);
    #else
    CORRADE_ASSERT_UNREACHABLE();
    #endif
    return result;
}
#endif
#endif

void AbstractQuery::begin() {
    #ifdef MAGNUM_BUILD_DEPRECATED
    CORRADE_INTERNAL_ASSERT(_target);
    #endif

    #ifndef MAGNUM_TARGET_GLES2
    glBeginQuery(_target, _id);
    #elif !defined(CORRADE_TARGET_EMSCRIPTEN)
    glBeginQueryEXT(_target, _id);
    #else
    CORRADE_ASSERT_UNREACHABLE();
    #endif
}

#ifdef MAGNUM_BUILD_DEPRECATED
void AbstractQuery::begin(const GLenum target) {
    CORRADE_INTERNAL_ASSERT(!_target || _target == target);

    _target = target;
    begin();
}
#endif

void AbstractQuery::end() {
    #ifdef MAGNUM_BUILD_DEPRECATED
    CORRADE_INTERNAL_ASSERT(_target);
    #endif

    #ifndef MAGNUM_TARGET_GLES2
    glEndQuery(_target);
    #elif !defined(CORRADE_TARGET_EMSCRIPTEN)
    glEndQueryEXT(_target);
    #else
    CORRADE_ASSERT_UNREACHABLE();
    #endif
}

}
